@@ -53,7 +53,7 @@ class Frontend(object):
         :type instruction_form: `dict`
         :returns: `True` if comment line, `False` otherwise
         """
-        return instruction_form["comment"] is not None and instruction_form["instruction"] is None
+        return instruction_form.comment is not None and instruction_form.mnemonic is None
 
     def throughput_analysis(self, kernel, show_lineno=False, show_cmnts=True):
         """
@@ -80,15 +80,19 @@ class Frontend(object):
         s += lineno_filler + self._get_port_number_line(port_len) + "\n"
         s += separator + "\n"
         for instruction_form in kernel:
+            if KernelDG.is_load_line_number(instruction_form.line_number):
+                continue
             line = "{:4d} {} {} {}".format(
-                instruction_form["line_number"],
+                instruction_form.line_number,
                 self._get_port_pressure(
-                    instruction_form["port_pressure"], port_len, separator=sep_list
+                    instruction_form.port_pressure, port_len, separator=sep_list
                 ),
-                self._get_flag_symbols(instruction_form["flags"])
-                if instruction_form["instruction"] is not None
-                else " ",
-                instruction_form["line"].strip().replace("\t", " "),
+                (
+                    self._get_flag_symbols(instruction_form.flags)
+                    if instruction_form.mnemonic is not None
+                    else " "
+                ),
+                instruction_form.line.strip().replace("\t", " "),
             )
             line = line if show_lineno else col_sep + col_sep.join(line.split(col_sep)[1:])
             if show_cmnts is False and self._is_comment(instruction_form):
@@ -110,22 +114,24 @@ class Frontend(object):
         """
         s = "\n\nLatency Analysis Report\n-----------------------\n"
         for instruction_form in cp_kernel:
+            if KernelDG.is_load_line_number(instruction_form.line_number):
+                continue
             s += (
                 "{:4d} {} {:4.1f} {}{}{} {}".format(
-                    instruction_form["line_number"],
+                    instruction_form.line_number,
                     separator,
-                    instruction_form["latency_cp"],
+                    instruction_form.latency_cp,
                     separator,
-                    "X" if INSTR_FLAGS.LT_UNKWN in instruction_form["flags"] else " ",
+                    "X" if INSTR_FLAGS.LT_UNKWN in instruction_form.flags else " ",
                     separator,
-                    instruction_form["line"],
+                    instruction_form.line,
                 )
             ) + "\n"
         s += (
             "\n{:4} {} {:4.1f}".format(
-                " " * max([len(str(instr_form["line_number"])) for instr_form in cp_kernel]),
+                " " * max([len(str(instr_form.line_number)) for instr_form in cp_kernel]),
                 " " * len(separator),
-                sum([instr_form["latency_cp"] for instr_form in cp_kernel]),
+                sum([instr_form.latency_cp for instr_form in cp_kernel]),
             )
         ) + "\n"
         return s
@@ -145,14 +151,17 @@ class Frontend(object):
         )
         # TODO find a way to overcome padding for different tab-lengths
         for dep in sorted(dep_dict.keys()):
-            s += "{:4d} {} {:4.1f} {} {:36}{} {}\n".format(
-                int(dep.split("-")[0]),
+            dep0 = float(dep.split("-")[0])
+            if KernelDG.is_load_line_number(dep0):
+                continue
+            s += "{:4.0f} {} {:4.1f} {} {:36}{} {}\n".format(
+                dep0,
                 separator,
                 dep_dict[dep]["latency"],
                 separator,
-                dep_dict[dep]["root"]["line"].strip(),
+                dep_dict[dep]["root"].line.strip(),
                 separator,
-                [node["line_number"] for node, lat in dep_dict[dep]["dependencies"]],
+                [node.line_number for node, lat in dep_dict[dep]["dependencies"]],
             )
         return s
 
@@ -201,6 +210,99 @@ class Frontend(object):
             + self.loopcarried_dependencies(kernel_dg.get_loopcarried_dependencies())
         )
 
+    def full_analysis_dict(
+        self,
+        kernel,
+        kernel_dg: KernelDG,
+        arch_warning=False,
+        length_warning=False,
+        lcd_warning=False,
+    ):
+        """
+        Create a dictionary of the full analysis for machine-readable output.
+
+        :param kernel: kernel to report on
+        :type kernel: list
+        :param kernel_dg: directed graph containing CP and LCD
+        :type kernel_dg: :class:`~osaca.semantics.KernelDG`
+        :param arch_warning: flag for additional user warning to specify micro-arch
+        :type arch_warning: boolean, optional
+        :param length_warning: flag for additional user warning to specify kernel length with
+                                     --lines
+        :type length_warning: boolean, optional
+        :param lcd_warning: flag for additional user warning due to LCD analysis timed out
+        :type lcd_warning: boolean, optional
+
+        :returns: dict -- a dict of the analysis
+        """
+        warnings = []
+
+        if arch_warning:
+            warnings.append("ArchWarning")
+
+        if length_warning:
+            warnings.append("LengthWarning")
+
+        if lcd_warning:
+            warnings.append("LCDWarning")
+
+        if INSTR_FLAGS.TP_UNKWN in [flag for instr in kernel for flag in instr.flags]:
+            warnings.append("UnknownInstrWarning")
+
+        tp_sum = ArchSemantics.get_throughput_sum(kernel) or kernel[0].port_pressure
+        cp_kernel = kernel_dg.get_critical_path()
+
+        dep_dict = kernel_dg.get_loopcarried_dependencies()
+        lcd_sum = 0.0
+        if dep_dict:
+            longest_lcd = max(dep_dict, key=lambda ln: dep_dict[ln]["latency"])
+            lcd_sum = dep_dict[longest_lcd]["latency"]
+        return {
+            "Header": self._header_report_dict(),
+            "Warnings": warnings,
+            "Kernel": [
+                {
+                    "Line": re.sub(r"\s+", " ", x.line.strip()),
+                    "LineNumber": x.line_number,
+                    "Flags": list(x.flags),
+                    "Instruction": x.mnemonic,
+                    "Operands": x.operands,
+                    "SemanticOperands": x.semantic_operands,
+                    "Label": x.label,
+                    "Directive": x.directive,
+                    "Latency": float(x.latency),
+                    "LatencyCP": float(x.latency_cp),
+                    "LatencyLCD": float(x.latency_lcd),
+                    "Throughput": float(x.throughput),
+                    "LatencyWithoutLoad": float(x.latency_wo_load),
+                    "PortPressure": {
+                        self._machine_model.get_ports()[i]: v
+                        for i, v in enumerate(x.port_pressure)
+                    },
+                    "PortUops": [
+                        {
+                            "Ports": list(y[1]),
+                            "Cycles": y[0],
+                        }
+                        for y in x.port_uops
+                    ],
+                    "Comment": x.comment,
+                }
+                for x in kernel
+            ],
+            "Summary": {
+                "PortPressure": {
+                    self._machine_model.get_ports()[i]: v for i, v in enumerate(tp_sum)
+                },
+                "CriticalPath": sum([x.latency_cp for x in cp_kernel]),
+                "LCD": lcd_sum,
+            },
+            "Target": {
+                "Name": self._arch.upper(),
+                "Ports": list(self._machine_model.get_ports()),
+            },
+        }
+
     def combined_view(
         self,
         kernel,
@@ -231,7 +333,7 @@ class Frontend(object):
         # Separator for ports
         separator = "-" * sum([x + 3 for x in port_len]) + "-"
         # ... for line numbers
-        separator += "--" + len(str(kernel[-1]["line_number"])) * "-"
+        separator += "--" + len(str(kernel[-1].line_number)) * "-"
         col_sep = "|"
         # for LCD/CP column
         separator += "-" * (2 * 6 + len(col_sep)) + "-" * len(col_sep) + "--"
@@ -239,14 +341,14 @@ class Frontend(object):
         headline = "Port pressure in cycles"
         headline_str = "{{:^{}}}".format(len(separator))
         # Prepare CP/LCD variable
-        cp_lines = [x["line_number"] for x in cp_kernel]
+        cp_lines = [x.line_number for x in cp_kernel]
         lcd_sum = 0.0
         lcd_lines = {}
         if dep_dict:
             longest_lcd = max(dep_dict, key=lambda ln: dep_dict[ln]["latency"])
             lcd_sum = dep_dict[longest_lcd]["latency"]
             lcd_lines = {
-                instr["line_number"]: lat for instr, lat in dep_dict[longest_lcd]["dependencies"]
+                instr.line_number: lat for instr, lat in dep_dict[longest_lcd]["dependencies"]
             }
 
         port_line = (
@@ -260,31 +362,35 @@ class Frontend(object):
         for instruction_form in kernel:
             if show_cmnts is False and self._is_comment(instruction_form):
                 continue
-            line_number = instruction_form["line_number"]
-            used_ports = [list(uops[1]) for uops in instruction_form["port_uops"]]
+            line_number = instruction_form.line_number
+            if KernelDG.is_load_line_number(line_number):
+                continue
+            used_ports = [list(uops[1]) for uops in instruction_form.port_uops]
             used_ports = list(set([p for uops_ports in used_ports for p in uops_ports]))
             s += "{:4d} {}{} {} {}\n".format(
                 line_number,
                 self._get_port_pressure(
-                    instruction_form["port_pressure"], port_len, used_ports, sep_list
+                    instruction_form.port_pressure, port_len, used_ports, sep_list
                 ),
                 self._get_lcd_cp_ports(
-                    instruction_form["line_number"],
+                    instruction_form.line_number,
                     cp_kernel if line_number in cp_lines else None,
                     lcd_lines.get(line_number),
                 ),
-                self._get_flag_symbols(instruction_form["flags"])
-                if instruction_form["instruction"] is not None
-                else " ",
-                instruction_form["line"].strip().replace("\t", " "),
+                (
+                    self._get_flag_symbols(instruction_form.flags)
+                    if instruction_form.mnemonic is not None
+                    else " "
+                ),
+                instruction_form.line.strip().replace("\t", " "),
             )
         s += "\n"
         # check for unknown instructions and throw warning if called without --ignore-unknown
         if not ignore_unknown and INSTR_FLAGS.TP_UNKWN in [
-            flag for instr in kernel for flag in instr["flags"]
+            flag for instr in kernel for flag in instr.flags
         ]:
             num_missing = len(
-                [instr["flags"] for instr in kernel if INSTR_FLAGS.TP_UNKWN in instr["flags"]]
+                [instr.flags for instr in kernel if INSTR_FLAGS.TP_UNKWN in instr.flags]
             )
             s += self._missing_instruction_error(num_missing)
         else:
@@ -292,8 +398,8 @@ class Frontend(object):
             tp_sum = ArchSemantics.get_throughput_sum(kernel)
             # if ALL instructions are unknown, take a line of 0s
             if not tp_sum:
-                tp_sum = kernel[0]["port_pressure"]
-            cp_sum = sum([x["latency_cp"] for x in cp_kernel])
+                tp_sum = kernel[0].port_pressure
+            cp_sum = sum([x.latency_cp for x in cp_kernel])
             s += (
                 lineno_filler
                 + self._get_port_pressure(tp_sum, port_len, separator=" ")
@@ -406,14 +512,14 @@ class Frontend(object):
 
     def _get_node_by_lineno(self, lineno, kernel):
         """Returns instruction form from kernel by its line number."""
-        nodes = [instr for instr in kernel if instr["line_number"] == lineno]
+        nodes = [instr for instr in kernel if instr.line_number == lineno]
         return nodes[0] if len(nodes) > 0 else None
 
     def _get_lcd_cp_ports(self, line_number, cp_dg, dep_lat, separator="|"):
         """Returns the CP and LCD line for one instruction."""
         lat_cp = lat_lcd = ""
         if cp_dg:
-            lat_cp = float(self._get_node_by_lineno(line_number, cp_dg)["latency_cp"])
+            lat_cp = float(self._get_node_by_lineno(line_number, cp_dg).latency_cp)
         if dep_lat is not None:
             lat_lcd = float(dep_lat)
         return "{} {:>4} {} {:>4} {}".format(separator, lat_cp, separator, lat_lcd, separator)
@@ -422,7 +528,7 @@ class Frontend(object):
         """Returns the maximal length needed to print all throughputs of the kernel."""
         port_len = [4 for x in self._machine_model.get_ports()]
         for instruction_form in kernel:
-            for i, port in enumerate(instruction_form["port_pressure"]):
+            for i, port in enumerate(instruction_form.port_pressure):
                 if len("{:.2f}".format(port)) > port_len[i]:
                     port_len[i] = len("{:.2f}".format(port))
         return port_len
@@ -449,6 +555,15 @@ class Frontend(object):
         )
         return header + "\n"
 
+    def _header_report_dict(self):
+        """Return header information in a dictionary format"""
+        return {
+            "Version": _get_version("__init__.py"),
+            "FileName": self._filename,
+            "Architecture": self._arch,
+            "Timestamp": dt.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
+        }
+
     def _symbol_map(self):
         """Prints instruction flag map."""
         symbol_dict = {
