@@ -1,20 +1,29 @@
 #!/usr/bin/env python3
 
 import copy
-import os
-import signal
+from enum import Enum
 import time
-from itertools import chain
+from itertools import chain, groupby
 from multiprocessing import Manager, Process, cpu_count
 
 import networkx as nx
 from osaca.semantics import INSTR_FLAGS, ArchSemantics, MachineModel
+from osaca.parser.instruction_form import InstructionForm
+from osaca.parser.memory import MemoryOperand
+from osaca.parser.register import RegisterOperand
+from osaca.parser.immediate import ImmediateOperand
+from osaca.parser.flag import FlagOperand
 
 
 class KernelDG(nx.DiGraph):
     # threshold for checking dependency graph sequential or in parallel
     INSTRUCTION_THRESHOLD = 50
 
+    class ReadKind(Enum):
+        NOT_A_READ = 0
+        READ_FOR_LOAD = 1
+        OTHER_READ = 2
+
     def __init__(
         self,
         parsed_kernel,
@@ -34,7 +43,8 @@ class KernelDG(nx.DiGraph):
             self.kernel, timeout, flag_dependencies
         )
 
-    def _extend_path(self, dst_list, kernel, dg, offset):
+    @classmethod
+    def _extend_path(cls, dst_list, kernel, dg, offset):
         for instr in kernel:
             generator_path = nx.algorithms.simple_paths.all_simple_paths(
                 dg, instr.line_number, instr.line_number + offset
@@ -43,6 +53,23 @@ class KernelDG(nx.DiGraph):
             dst_list.extend(tmp_list)
         # print('Thread [{}-{}] done'.format(kernel[0]['line_number'], kernel[-1]['line_number']))
 
+    @staticmethod
+    def get_load_line_number(line_number):
+        # The line number of the load must be less than the line number of the instruction.  The
+        # offset is irrelevant, but it must be a machine number with trailing zeroes to avoid silly
+        # rounding issues.
+        return line_number - 0.125
+
+    @staticmethod
+    def is_load_line_number(line_number):
+        return line_number != int(line_number)
+
+    @staticmethod
+    def get_real_line_number(line_number):
+        return (
+            int(line_number + 0.125) if KernelDG.is_load_line_number(line_number) else line_number
+        )
+
     def create_DG(self, kernel, flag_dependencies=False):
         """
         Create directed graph from given kernel
@@ -54,48 +81,62 @@ class KernelDG(nx.DiGraph):
         :type flag_dependencies: boolean, optional
         :returns: :class:`~nx.DiGraph` -- directed graph object
         """
-        # 1. go through kernel instruction forms and add them as node attribute
-        # 2. find edges (to dependend further instruction)
-        # 3. get LT value and set as edge weight
+        # Go through kernel instruction forms and add them as nodes of the graph.  Create a LOAD
+        # node for instructions that include a memory reference.
         dg = nx.DiGraph()
+        loads = {}
         for i, instruction_form in enumerate(kernel):
-            dg.add_node(instruction_form["line_number"])
-            dg.nodes[instruction_form["line_number"]]["instruction_form"] = instruction_form
+            dg.add_node(instruction_form.line_number)
+            dg.nodes[instruction_form.line_number]["instruction_form"] = instruction_form
             # add load as separate node if existent
             if (
-                INSTR_FLAGS.HAS_LD in instruction_form["flags"]
-                and INSTR_FLAGS.LD not in instruction_form["flags"]
+                INSTR_FLAGS.HAS_LD in instruction_form.flags
+                and INSTR_FLAGS.LD not in instruction_form.flags
             ):
                 # add new node
-                dg.add_node(instruction_form["line_number"] + 0.1)
-                dg.nodes[instruction_form["line_number"] + 0.1][
-                    "instruction_form"
-                ] = instruction_form
+                load_line_number = KernelDG.get_load_line_number(instruction_form.line_number)
+                loads[instruction_form.line_number] = load_line_number
+                dg.add_node(load_line_number)
+                dg.nodes[load_line_number]["instruction_form"] = InstructionForm(
+                    mnemonic="_LOAD_", line=instruction_form.line, line_number=load_line_number
+                )
                 # and set LD latency as edge weight
                 dg.add_edge(
-                    instruction_form["line_number"] + 0.1,
-                    instruction_form["line_number"],
-                    latency=instruction_form["latency"] - instruction_form["latency_wo_load"],
+                    load_line_number,
+                    instruction_form.line_number,
+                    latency=instruction_form.latency - instruction_form.latency_wo_load,
                 )
+
+        # 1. find edges (to dependend further instruction)
+        # 2. get LT value and set as edge weight
+        for i, instruction_form in enumerate(kernel):
             for dep, dep_flags in self.find_depending(
                 instruction_form, kernel[i + 1 :], flag_dependencies
             ):
+                # print(instruction_form.line_number,"\t",dep.line_number,"\n")
                 edge_weight = (
-                    instruction_form["latency"]
-                    if "mem_dep" in dep_flags or "latency_wo_load" not in instruction_form
-                    else instruction_form["latency_wo_load"]
+                    instruction_form.latency
+                    if "mem_dep" in dep_flags or instruction_form.latency_wo_load is None
+                    else instruction_form.latency_wo_load
                 )
                 if "storeload_dep" in dep_flags and self.model is not None:
                     edge_weight += self.model.get("store_to_load_forward_latency", 0)
                 if "p_indexed" in dep_flags and self.model is not None:
                     edge_weight = self.model.get("p_index_latency", 1)
-                dg.add_edge(
-                    instruction_form["line_number"],
-                    dep["line_number"],
-                    latency=edge_weight,
-                )
+                if "for_load" in dep_flags and self.model is not None and dep.line_number in loads:
+                    dg.add_edge(
+                        instruction_form.line_number,
+                        loads[dep.line_number],
+                        latency=edge_weight,
+                    )
+                else:
+                    dg.add_edge(
+                        instruction_form.line_number,
+                        dep.line_number,
+                        latency=edge_weight,
+                    )
 
-                dg.nodes[dep["line_number"]]["instruction_form"] = dep
+                dg.nodes[dep.line_number]["instruction_form"] = dep
         return dg
 
     def check_for_loopcarried_dep(self, kernel, timeout=10, flag_dependencies=False):
@@ -114,7 +155,7 @@ class KernelDG(nx.DiGraph):
         tmp_kernel = [] + kernel
         for orig_iform in kernel:
             temp_iform = copy.copy(orig_iform)
-            temp_iform["line_number"] += offset
+            temp_iform.line_number += offset
             tmp_kernel.append(temp_iform)
         # get dependency graph
         dg = self.create_DG(tmp_kernel, flag_dependencies)
@@ -135,7 +176,7 @@ class KernelDG(nx.DiGraph):
                 all_paths = manager.list()
                 processes = [
                     Process(
-                        target=self._extend_path,
+                        target=KernelDG._extend_path,
                         args=(all_paths, instr_section, dg, offset),
                     )
                     for instr_section in instrs
@@ -161,9 +202,7 @@ class KernelDG(nx.DiGraph):
                         # terminate running processes
                         for p in processes:
                             if p.is_alive():
-                                # Python 3.6 does not support Process.kill().
-                                # Can be changed to `p.kill()` after EoL (01/22) of Py3.6
-                                os.kill(p.pid, signal.SIGKILL)
+                                p.kill()
                             p.join()
                 all_paths = list(all_paths)
         else:
@@ -183,11 +222,11 @@ class KernelDG(nx.DiGraph):
             for s, d in nx.utils.pairwise(path):
                 edge_lat = dg.edges[s, d]["latency"]
                 # map source node back to original line numbers
-                if s >= offset:
+                if s > offset:
                     s -= offset
                 lat_path.append((s, edge_lat))
                 lat_sum += edge_lat
-            if d >= offset:
+            if d > offset:
                 d -= offset
             lat_path.sort()
 
@@ -204,46 +243,40 @@ class KernelDG(nx.DiGraph):
         for lat_sum, involved_lines in loopcarried_deps:
             dict_key = "-".join([str(il[0]) for il in involved_lines])
             loopcarried_deps_dict[dict_key] = {
-                "root": self._get_node_by_lineno(involved_lines[0][0]),
+                "root": self._get_node_by_lineno(dg, involved_lines[0][0]),
                 "dependencies": [
-                    (self._get_node_by_lineno(ln), lat) for ln, lat in involved_lines
+                    (self._get_node_by_lineno(dg, ln), lat) for ln, lat in involved_lines
                 ],
                 "latency": lat_sum,
             }
         return loopcarried_deps_dict
 
-    def _get_node_by_lineno(self, lineno, kernel=None, all=False):
-        """Return instruction form with line number ``lineno`` from  kernel"""
-        if kernel is None:
-            kernel = self.kernel
-        result = [instr for instr in kernel if instr.line_number == lineno]
-        if not all:
-            return result[0]
-        else:
-            return result
+    def _get_node_by_lineno(self, dg, lineno):
+        """Return instruction form with line number ``lineno`` from  dg"""
+        return dg.nodes[lineno]["instruction_form"]
 
     def get_critical_path(self):
         """Find and return critical path after the creation of a directed graph."""
-        max_latency_instr = max(self.kernel, key=lambda k: k["latency"])
+        max_latency_instr = max(self.kernel, key=lambda k: k.latency)
         if nx.algorithms.dag.is_directed_acyclic_graph(self.dg):
             longest_path = nx.algorithms.dag.dag_longest_path(self.dg, weight="latency")
             # TODO verify that we can remove the next two lince due to earlier initialization
             for line_number in longest_path:
-                self._get_node_by_lineno(int(line_number))["latency_cp"] = 0
+                self._get_node_by_lineno(self.dg, line_number).latency_cp = 0
             # set cp latency to instruction
             path_latency = 0.0
             for s, d in nx.utils.pairwise(longest_path):
-                node = self._get_node_by_lineno(int(s))
-                node["latency_cp"] = self.dg.edges[(s, d)]["latency"]
-                path_latency += node["latency_cp"]
+                node = self._get_node_by_lineno(self.dg, s)
+                node.latency_cp = self.dg.edges[(s, d)]["latency"]
+                path_latency += node.latency_cp
             # add latency for last instruction
-            node = self._get_node_by_lineno(int(longest_path[-1]))
-            node["latency_cp"] = node["latency"]
-            if max_latency_instr["latency"] > path_latency:
-                max_latency_instr["latency_cp"] = float(max_latency_instr["latency"])
+            node = self._get_node_by_lineno(self.dg, longest_path[-1])
+            node.latency_cp = node.latency
+            if max_latency_instr.latency > path_latency:
+                max_latency_instr.latency_cp = float(max_latency_instr.latency)
                 return [max_latency_instr]
             else:
-                return [x for x in self.kernel if x["line_number"] in longest_path]
+                return [self._get_node_by_lineno(self.dg, x) for x in longest_path]
         else:
             # split to DAG
             raise NotImplementedError("Kernel is cyclic.")
@@ -272,8 +305,8 @@ class KernelDG(nx.DiGraph):
         if instruction_form.semantic_operands is None:
             return
         for dst in chain(
-            instruction_form.semantic_operands.destination,
-            instruction_form.semantic_operands.src_dst,
+            instruction_form.semantic_operands["destination"],
+            instruction_form.semantic_operands["src_dst"],
         ):
             # TODO instructions before must be considered as well, if they update registers
             # not used by insruction_form. E.g., validation/build/A64FX/gcc/O1/gs-2d-5pt.marked.s
@@ -282,27 +315,34 @@ class KernelDG(nx.DiGraph):
             for i, instr_form in enumerate(instructions):
                 self._update_reg_changes(instr_form, register_changes)
                 # print("  TO", instr_form.line, register_changes)
-                if "register" in dst:
+                if isinstance(dst, RegisterOperand):
                     # read of register
-                    if self.is_read(dst.register, instr_form):
-                        if dst.get("pre_indexed", False) or dst.get("post_indexed", False):
-                            yield instr_form, ["p_indexed"]
-                        else:
-                            yield instr_form, []
+                    read_kind = self._read_kind(dst, instr_form)
+                    if read_kind != KernelDG.ReadKind.NOT_A_READ:
+                        dep_flags = []
+                        if (
+                            dst.pre_indexed
+                            or dst.post_indexed
+                            or (isinstance(dst.post_indexed, dict))
+                        ):
+                            dep_flags = ["p_indexed"]
+                        if read_kind == KernelDG.ReadKind.READ_FOR_LOAD:
+                            dep_flags += ["for_load"]
+                        yield instr_form, dep_flags
                     # write to register -> abort
-                    if self.is_written(dst.register, instr_form):
+                    if self.is_written(dst, instr_form):
                         break
-                if "flag" in dst and flag_dependencies:
+                if isinstance(dst, FlagOperand) and flag_dependencies:
                     # read of flag
-                    if self.is_read(dst.flag, instr_form):
+                    if self.is_read(dst, instr_form):
                         yield instr_form, []
                     # write to flag -> abort
-                    if self.is_written(dst.flag, instr_form):
+                    if self.is_written(dst, instr_form):
                         break
-                if "memory" in dst:
+                if isinstance(dst, MemoryOperand):
                     # base register is altered during memory access
-                    if "pre_indexed" in dst.memory:
-                        if self.is_written(dst.memory.base, instr_form):
+                    if dst.pre_indexed:
+                        if self.is_written(dst.base, instr_form):
                             break
                     # if dst.memory.base:
                     #    if self.is_read(dst.memory.base, instr_form):
@@ -310,18 +350,18 @@ class KernelDG(nx.DiGraph):
                     # if dst.memory.index:
                     #    if self.is_read(dst.memory.index, instr_form):
                     #        yield instr_form, []
-                    if "post_indexed" in dst.memory:
+                    if dst.post_indexed:
                         # Check for read of base register until overwrite
-                        if self.is_written(dst.memory.base, instr_form):
+                        if self.is_written(dst.base, instr_form):
                             break
                     # TODO record register changes
                     #      (e.g., mov, leaadd, sub, inc, dec) in instructions[:i]
                     #      and pass to is_memload and is_memstore to consider relevance.
                     # load from same location (presumed)
-                    if self.is_memload(dst.memory, instr_form, register_changes):
+                    if self.is_memload(dst, instr_form, register_changes):
                         yield instr_form, ["storeload_dep"]
                     # store to same location (presumed)
-                    if self.is_memstore(dst.memory, instr_form, register_changes):
+                    if self.is_memstore(dst, instr_form, register_changes):
                         break
                 self._update_reg_changes(instr_form, register_changes, only_postindexed=True)
 
@@ -359,68 +399,87 @@ class KernelDG(nx.DiGraph):
             return self.dg.successors(line_number)
         return iter([])
 
-    def is_read(self, register, instruction_form):
-        """Check if instruction form reads from given register"""
+    def _read_kind(self, register, instruction_form):
+        """Check if instruction form reads from given register.  Returns a ReadKind."""
         is_read = False
+        for_load = False
         if instruction_form.semantic_operands is None:
-            return is_read
+            return KernelDG.ReadKind.NOT_A_READ
         for src in chain(
-            instruction_form.semantic_operands.source,
-            instruction_form.semantic_operands.src_dst,
+            instruction_form.semantic_operands["source"],
+            instruction_form.semantic_operands["src_dst"],
         ):
-            if "register" in src:
-                is_read = self.parser.is_reg_dependend_of(register, src.register) or is_read
-            if "flag" in src:
-                is_read = self.parser.is_flag_dependend_of(register, src.flag) or is_read
-            if "memory" in src:
-                if src.memory.base is not None:
-                    is_read = self.parser.is_reg_dependend_of(register, src.memory.base) or is_read
-                if src.memory.index is not None:
-                    is_read = (
-                        self.parser.is_reg_dependend_of(register, src.memory.index) or is_read
+            if isinstance(src, RegisterOperand):
+                is_read = self.parser.is_reg_dependend_of(register, src) or is_read
+            if isinstance(src, FlagOperand):
+                is_read = self.parser.is_flag_dependend_of(register, src) or is_read
+            if isinstance(src, MemoryOperand):
+                is_memory_read = False
+                if src.base is not None:
+                    is_memory_read = self.parser.is_reg_dependend_of(register, src.base)
+                if src.index is not None and isinstance(src.index, RegisterOperand):
+                    is_memory_read = (
+                        self.parser.is_reg_dependend_of(register, src.index) or is_memory_read
                     )
+                for_load = is_memory_read
+                is_read = is_read or is_memory_read
         # Check also if read in destination memory address
         for dst in chain(
-            instruction_form.semantic_operands.destination,
-            instruction_form.semantic_operands.src_dst,
+            instruction_form.semantic_operands["destination"],
+            instruction_form.semantic_operands["src_dst"],
         ):
-            if "memory" in dst:
-                if dst.memory.base is not None:
-                    is_read = self.parser.is_reg_dependend_of(register, dst.memory.base) or is_read
-                if dst.memory.index is not None:
-                    is_read = (
-                        self.parser.is_reg_dependend_of(register, dst.memory.index) or is_read
-                    )
-        return is_read
+            if isinstance(dst, MemoryOperand):
+                if dst.base is not None:
+                    is_read = self.parser.is_reg_dependend_of(register, dst.base) or is_read
+                if dst.index is not None:
+                    is_read = self.parser.is_reg_dependend_of(register, dst.index) or is_read
+        if is_read:
+            if for_load:
+                return KernelDG.ReadKind.READ_FOR_LOAD
+            else:
+                return KernelDG.ReadKind.OTHER_READ
+        else:
+            return KernelDG.ReadKind.NOT_A_READ
+
+    def is_read(self, register, instruction_form):
+        return self._read_kind(register, instruction_form) != KernelDG.ReadKind.NOT_A_READ
 
     def is_memload(self, mem, instruction_form, register_changes={}):
         """Check if instruction form loads from given location, assuming register_changes"""
         if instruction_form.semantic_operands is None:
             return False
         for src in chain(
-            instruction_form.semantic_operands.source,
-            instruction_form.semantic_operands.src_dst,
+            instruction_form.semantic_operands["source"],
+            instruction_form.semantic_operands["src_dst"],
         ):
             # Here we check for mem dependecies only
-            if "memory" not in src:
+            if not isinstance(src, MemoryOperand):
                 continue
-            src = src.memory
+            # src = src.memory
 
             # determine absolute address change
             addr_change = 0
-            if src.offset and "value" in src.offset:
+            if isinstance(src.offset, ImmediateOperand) and src.offset.value is not None:
                 addr_change += src.offset.value
-            if mem.offset:
+            if isinstance(mem.offset, ImmediateOperand) and mem.offset.value is not None:
                 addr_change -= mem.offset.value
             if mem.base and src.base:
                 base_change = register_changes.get(
-                    src.base.get("prefix", "") + src.base.name,
-                    {"name": src.base.get("prefix", "") + src.base.name, "value": 0},
+                    (src.base.prefix if src.base.prefix is not None else "") + src.base.name,
+                    {
+                        "name": (src.base.prefix if src.base.prefix is not None else "")
+                        + src.base.name,
+                        "value": 0,
+                    },
                 )
                 if base_change is None:
                     # Unknown change occurred
                     continue
-                if mem.base.get("prefix", "") + mem.base["name"] != base_change["name"]:
+                if (
+                    mem.base.prefix
+                    if mem.base.prefix is not None
+                    else "" + mem.base.name != base_change["name"]
+                ):
                     # base registers do not match
                     continue
                 addr_change += base_change["value"]
@@ -429,8 +488,12 @@ class KernelDG(nx.DiGraph):
                 continue
             if mem.index and src.index:
                 index_change = register_changes.get(
-                    src.index.get("prefix", "") + src.index.name,
-                    {"name": src.index.get("prefix", "") + src.index.name, "value": 0},
+                    (src.index.prefix if src.index.prefix is not None else "") + src.index.name,
+                    {
+                        "name": (src.index.prefix if src.index.prefix is not None else "")
+                        + src.index.name,
+                        "value": 0,
+                    },
                 )
                 if index_change is None:
                     # Unknown change occurred
@@ -438,7 +501,11 @@ class KernelDG(nx.DiGraph):
                 if mem.scale != src.scale:
                     # scale factors do not match
                     continue
-                if mem.index.get("prefix", "") + mem.index["name"] != index_change["name"]:
+                if (
+                    mem.index.prefix
+                    if mem.index.prefix is not None
+                    else "" + mem.index.name != index_change["name"]
+                ):
                     # index registers do not match
                     continue
                 addr_change += index_change["value"] * src.scale
@@ -456,28 +523,24 @@ class KernelDG(nx.DiGraph):
         if instruction_form.semantic_operands is None:
             return is_written
         for dst in chain(
-            instruction_form.semantic_operands.destination,
-            instruction_form.semantic_operands.src_dst,
+            instruction_form.semantic_operands["destination"],
+            instruction_form.semantic_operands["src_dst"],
         ):
-            if "register" in dst:
-                is_written = self.parser.is_reg_dependend_of(register, dst.register) or is_written
-            if "flag" in dst:
-                is_written = self.parser.is_flag_dependend_of(register, dst.flag) or is_written
-            if "memory" in dst:
-                if "pre_indexed" in dst.memory or "post_indexed" in dst.memory:
-                    is_written = (
-                        self.parser.is_reg_dependend_of(register, dst.memory.base) or is_written
-                    )
+            if isinstance(dst, RegisterOperand):
+                is_written = self.parser.is_reg_dependend_of(register, dst) or is_written
+            if isinstance(dst, FlagOperand):
+                is_written = self.parser.is_flag_dependend_of(register, dst) or is_written
+            if isinstance(dst, MemoryOperand):
+                if dst.pre_indexed or dst.post_indexed:
+                    is_written = self.parser.is_reg_dependend_of(register, dst.base) or is_written
         # Check also for possible pre- or post-indexing in memory addresses
         for src in chain(
-            instruction_form.semantic_operands.source,
-            instruction_form.semantic_operands.src_dst,
+            instruction_form.semantic_operands["source"],
+            instruction_form.semantic_operands["src_dst"],
         ):
-            if "memory" in src:
-                if "pre_indexed" in src.memory or "post_indexed" in src.memory:
-                    is_written = (
-                        self.parser.is_reg_dependend_of(register, src.memory.base) or is_written
-                    )
+            if isinstance(src, MemoryOperand):
+                if src.pre_indexed or src.post_indexed:
+                    is_written = self.parser.is_reg_dependend_of(register, src.base) or is_written
         return is_written
 
     def is_memstore(self, mem, instruction_form, register_changes={}):
@@ -486,11 +549,11 @@ class KernelDG(nx.DiGraph):
         if instruction_form.semantic_operands is None:
             return is_store
         for dst in chain(
-            instruction_form.semantic_operands.destination,
-            instruction_form.semantic_operands.src_dst,
+            instruction_form.semantic_operands["destination"],
+            instruction_form.semantic_operands["src_dst"],
         ):
-            if "memory" in dst:
-                is_store = mem == dst["memory"] or is_store
+            if isinstance(dst, MemoryOperand):
+                is_store = mem == dst or is_store
         return is_store
 
     def export_graph(self, filepath=None):
@@ -503,22 +566,19 @@ class KernelDG(nx.DiGraph):
         """
         graph = copy.deepcopy(self.dg)
         cp = self.get_critical_path()
-        cp_line_numbers = [x["line_number"] for x in cp]
+        cp_line_numbers = [x.line_number for x in cp]
         lcd = self.get_loopcarried_dependencies()
         lcd_line_numbers = {}
         for dep in lcd:
-            lcd_line_numbers[dep] = [x["line_number"] for x, lat in lcd[dep]["dependencies"]]
-        # add color scheme
-        graph.graph["node"] = {"colorscheme": "accent8"}
-        graph.graph["edge"] = {"colorscheme": "accent8"}
+            lcd_line_numbers[dep] = [x.line_number for x, lat in lcd[dep]["dependencies"]]
 
         # create LCD edges
         for dep in lcd_line_numbers:
             min_line_number = min(lcd_line_numbers[dep])
             max_line_number = max(lcd_line_numbers[dep])
-            graph.add_edge(max_line_number, min_line_number)
-            graph.edges[max_line_number, min_line_number]["latency"] = [
-                lat for x, lat in lcd[dep]["dependencies"] if x["line_number"] == max_line_number
+            graph.add_edge(min_line_number, max_line_number, dir="back")
+            graph.edges[min_line_number, max_line_number]["latency"] = [
+                lat for x, lat in lcd[dep]["dependencies"] if x.line_number == max_line_number
             ]
 
         # add label to edges
@@ -527,62 +587,83 @@ class KernelDG(nx.DiGraph):
 
         # add CP values to graph
         for n in cp:
-            graph.nodes[n["line_number"]]["instruction_form"]["latency_cp"] = n["latency_cp"]
+            graph.nodes[n.line_number]["instruction_form"].latency_cp = n.latency_cp
 
-        # color CP and LCD
+        # Make the critical path bold.
         for n in graph.nodes:
             if n in cp_line_numbers:
                 # graph.nodes[n]['color'] = 1
                 graph.nodes[n]["style"] = "bold"
                 graph.nodes[n]["penwidth"] = 4
-            for col, dep in enumerate(lcd):
-                if n in lcd_line_numbers[dep]:
-                    if "style" not in graph.nodes[n]:
-                        graph.nodes[n]["style"] = "filled"
-                    else:
-                        graph.nodes[n]["style"] += ",filled"
-                    graph.nodes[n]["fillcolor"] = 2 + col
 
-        # color edges
-        for e in graph.edges:
-            if (
-                graph.nodes[e[0]]["instruction_form"]["line_number"] in cp_line_numbers
-                and graph.nodes[e[1]]["instruction_form"]["line_number"] in cp_line_numbers
-                and e[0] < e[1]
+        # Make critical path edges bold.
+        for u, v in zip(cp_line_numbers[:-1], cp_line_numbers[1:]):
+            graph.edges[u, v]["style"] = "bold"
+            graph.edges[u, v]["penwidth"] = 3
+
+        # Color the cycles created by loop-carried dependencies, longest first, never recoloring
+        # any node or edge, so that the longest LCD and most long chains that are involved in the
+        # loop are legible.
+        lcd_by_latencies = sorted(
+            (
+                (latency, list(deps))
+                for latency, deps in groupby(lcd, lambda dep: lcd[dep]["latency"])
+            ),
+            reverse=True,
+        )
+        node_colors = {}
+        edge_colors = {}
+        colors_used = 0
+        for i, (latency, deps) in enumerate(lcd_by_latencies):
+            color = None
+            for dep in deps:
+                path = lcd_line_numbers[dep]
+                for n in path:
+                    if n not in node_colors:
+                        if not color:
+                            color = colors_used + 1
+                            colors_used += 1
+                        node_colors[n] = color
+                for u, v in zip(path, path[1:] + [path[0]]):
+                    if (u, v) not in edge_colors:
+                        # Don’t introduce a color just for an edge.
+                        if not color:
+                            color = colors_used
+                        edge_colors[u, v] = color
+        max_color = min(11, colors_used)
+        colorscheme = f"spectral{max(3, max_color)}"
+        graph.graph["node"] = {"colorscheme": colorscheme}
+        graph.graph["edge"] = {"colorscheme": colorscheme}
+        for n, color in node_colors.items():
+            if "style" not in graph.nodes[n]:
+                graph.nodes[n]["style"] = "filled"
+            else:
+                graph.nodes[n]["style"] += ",filled"
+            graph.nodes[n]["fillcolor"] = color
+            if (max_color >= 4 and color in (1, max_color)) or (
+                max_color >= 10 and color in (1, 2, max_color - 1, max_color)
             ):
-                bold_edge = True
-                for i in range(e[0] + 1, e[1]):
-                    if i in cp_line_numbers:
-                        bold_edge = False
-                if bold_edge:
-                    graph.edges[e]["style"] = "bold"
-                    graph.edges[e]["penwidth"] = 3
-            for dep in lcd_line_numbers:
-                if (
-                    graph.nodes[e[0]]["instruction_form"]["line_number"] in lcd_line_numbers[dep]
-                    and graph.nodes[e[1]]["instruction_form"]["line_number"]
-                    in lcd_line_numbers[dep]
-                ):
-                    graph.edges[e]["color"] = graph.nodes[e[1]]["fillcolor"]
+                graph.nodes[n]["fontcolor"] = "white"
+        for (u, v), color in edge_colors.items():
+            # The backward edge of the cycle is represented as the corresponding forward
+            # edge with the attribute dir=back.
+            edge = graph.edges[u, v] if (u, v) in graph.edges else graph.edges[v, u]
+            edge["color"] = color
 
         # rename node from [idx] to [idx mnemonic] and add shape
         mapping = {}
         for n in graph.nodes:
-            if int(n) != n:
-                mapping[n] = "{}: LOAD".format(int(n))
+            node = graph.nodes[n]["instruction_form"]
+            if node.mnemonic is not None:
+                mapping[n] = "{}: {}".format(KernelDG.get_real_line_number(n), node.mnemonic)
+            else:
+                label = "label" if node.label is not None else None
+                label = "directive" if node.directive is not None else label
+                label = "comment" if node.comment is not None and label is None else label
+                mapping[n] = "{}: {}".format(n, label)
                 graph.nodes[n]["fontname"] = "italic"
                 graph.nodes[n]["fontsize"] = 11.0
-            else:
-                node = graph.nodes[n]["instruction_form"]
-                if node["instruction"] is not None:
-                    mapping[n] = "{}: {}".format(n, node["instruction"])
-                else:
-                    label = "label" if node["label"] else None
-                    label = "directive" if node["directive"] else label
-                    label = "comment" if node["comment"] and label is None else label
-                    mapping[n] = "{}: {}".format(n, label)
-                    graph.nodes[n]["fontname"] = "italic"
-                    graph.nodes[n]["fontsize"] = 11.0
+            if not KernelDG.is_load_line_number(n):
                 graph.nodes[n]["shape"] = "rectangle"
 
         nx.relabel.relabel_nodes(graph, mapping, copy=False)
