@@ -6,7 +6,9 @@ Unit tests for base assembly parser
 import os
 import unittest
 
-from osaca.parser import AttrDict, BaseParser
+from osaca.parser import BaseParser
+from osaca.parser.register import RegisterOperand
+from osaca.parser.immediate import ImmediateOperand
 
 
 class TestBaseParser(unittest.TestCase):
@@ -18,6 +20,8 @@ class TestBaseParser(unittest.TestCase):
             pass
         with open(self._find_file("triad_x86_iaca.s")) as f:
             self.triad_code = f.read()
+        with open(self._find_file("triad_x86_intel.s")) as f:
+            self.triad_code_intel = f.read()
         with open(self._find_file("triad_arm_iaca.s")) as f:
             self.triad_code_arm = f.read()
         with open(self._find_file("kernel_x86.s")) as f:
@@ -44,8 +48,8 @@ class TestBaseParser(unittest.TestCase):
             self.parser.parse_instruction(instr1)
 
     def test_register_funcs(self):
-        reg_a1 = AttrDict({"name": "rax"})
-        reg_a2 = AttrDict({"name": "eax"})
+        reg_a1 = RegisterOperand(name="rax")
+        reg_a2 = RegisterOperand(name="eax")
         register_string = "v1.2d"
         with self.assertRaises(NotImplementedError):
             self.parser.is_reg_dependend_of(reg_a1, reg_a2)
@@ -61,15 +65,16 @@ class TestBaseParser(unittest.TestCase):
             self.parser.get_full_reg_name(reg_a1)
 
     def test_normalize_imd(self):
-        imd_hex_1 = {"value": "0x4f"}
+        imd_hex_1 = ImmediateOperand(value="0x4f")
         with self.assertRaises(NotImplementedError):
             self.parser.normalize_imd(imd_hex_1)
 
     def test_detect_ISA(self):
-        self.assertEqual(BaseParser.detect_ISA(self.triad_code), "x86")
-        self.assertEqual(BaseParser.detect_ISA(self.triad_code_arm), "aarch64")
-        self.assertEqual(BaseParser.detect_ISA(self.x86_code), "x86")
-        self.assertEqual(BaseParser.detect_ISA(self.aarch64_code), "aarch64")
+        self.assertEqual(BaseParser.detect_ISA(self.triad_code), ("x86", "ATT"))
+        self.assertEqual(BaseParser.detect_ISA(self.triad_code_intel), ("x86", "INTEL"))
+        self.assertEqual(BaseParser.detect_ISA(self.triad_code_arm), ("aarch64", None))
+        self.assertEqual(BaseParser.detect_ISA(self.x86_code), ("x86", "ATT"))
+        self.assertEqual(BaseParser.detect_ISA(self.aarch64_code), ("aarch64", None))
 
     ##################
     # Helper functions
